;Main idea stolen from http://www.siwawi.arubi.uni-kl.de/avr_projects/matrix44/
;Code & additional ideas (clock, stream) by Arne Rossius && KBinaryClock, 2005
;     ==> http://www.elektronik.de.vu <==> http://www.blinkenarea.org <==

;******************************************************************************
;**                                                                          **
;**   BB          LL            II        NN    NN    KK    KK    YY    YY   **
;**   BBBBBB      LL            II        NNNN  NN    KK  KK        YYYY     **
;**   BB    BB    LL            II        NN  NNNN    KKKKKK        YY       **
;**   BBBBBB      LLLLLLLL      II        NN    NN    KK    KK    YY         **
;**                                                                          **
;*********************************  BLINKY  ***********************************
;*********************************  -LAMPS  ***********************************
;**                                                                          **
;**               LL          AAAA        MM    MM    PPPPPP          SSSS   **
;**   --------    LL        AA    AA      MMMMMMMM    PP    PP      SS       **
;**               LL        AAAAAAAA      MM    MM    PPPPPP          SS     **
;**               LLLLLLLL  AA    AA      MM    MM    PP          SSSS       **
;**                                                                          **
;******************************************************************************

;ATTiny2313 @ 8MHz (internal R/C)

;recommended latches: 4x 74LS75 or 74HC75 and NPN output transistors
;connected to inverted-Q outputs

;FEATURES:
;  * play greyscale animations
;  * play RS232 stream (MCUF protocol)
;  * display (and set) BCD clock

;VERSION HISTORY
;  0.7  2006-03-08  First release of "BlinkyLamps" version for latched outputs
;

;PORTS:
;  PB0,1,6,7: rows (top ... bottom), active low, latch inputs (use Q' outputs!)
;  PB2,3,4,5: cols (right ... left), latch "STROBE" pins
;  PD0:       RS232 Rx
;  PD2:       left key (short: animations, long: stream, set clock mode: +1)
;  PD3:       right key (short: display clock, long: set clock, set clock mode: next digit/done)
;  PD4:       clock signal input (1 pulse per second)

;MATRIX:
;  +---+---+---+---+
;  | 3 | 7 | B | F |
;  +---+---+---+---+
;  | 2 | 6 | A | E |
;  +---+---+---+---+    LED mapping for animations:
;  | 1 | 5 | 9 | D |    0x3210, 0x7654, 0xBA98, 0xFEDC
;  +---+---+---+---+
;  | 0 | 4 | 8 | C |
;  +---+---+---+---+

.include "../include/tn2313def.inc"


.equ	Latches = 1 ;latches connected instead of LEDs
.equ	ClockPulses = 0 ;negative (0) or positive (1) 1Hz clock pulses
.equ	FREQ = 8000000 ;clock frequency
.equ	BAUD = 19200 ;RS232 BAUD rate
;multiplex frequency at 8MHz: approx. 160 Hz ==> 1 multiplex loop takes approx. 6.25ms
.equ	KEY_SHORT = 3 ;multiplex loops after which the key is consideres pressed
.equ	KEY_LONG = 192 ;multiplex loops after which the key is considered pressed long
.equ	RS232_TIMEOUT = 16 ;multiplex loops after which an inactive RS232 connection is timed out

.def	pwmcnt = R0
.def	mpxcnt = R1
.def	ani2L = R2
.def	ani2H = R3
.def	anistep = R4
.def	anisteps = R5
.def	aniold = R6
.def	aninew = R7
.def	aniloopL = R8
.def	aniloopH = R9
.def	aniloops = R10
.def	clkin = R11
.def	rs232cnt = R12

.def	temp = R16
.def	temp2 = R17
.def	mode = R18
.def	keycnt1 = R19
.def	keycnt2 = R20
.def	second = R21
.def	minute = R22
.def	hour = R23
.def	p_rs232 = R24
.def	aninum = R25
.def	ani1L = R26
.def	ani1H = R27
;R28/29 (Y) used for RAM access
;R30/31 (Z) used for flash access

;mode numbers
.equ	MODE_ANI = 0 ;play animations
.equ	MODE_STREAM_WAIT4RELEASE = 1 ;play RS232 stream, wait for release of the key first
.equ	MODE_STREAM = 2 ;play RS232 stream
.equ	MODE_CLOCK = 4 ;display BCD clock
.equ	MODE_CLOCKSET_WAIT4RELEASE = 5 ;set clock, wait for release of the key first
.equ	MODE_CLOCKSET = 6 ;set clock now

;animation commands:
.equ	ANI_END = 0x0000 ;ANI_END
.equ	CMD_SET = 0x1000 ;CMD_SET|Wait_Time, Col1, Col2, Col3, Col4 (1 <= Wait_Time <= 4095)
.equ	CMD_FADE = 0x2000 ;CMD_FADE|Steps, Col1, Col2, Col3, Col4, Step_Time (2 <= Steps <= 16)
.equ	CMD_LOOP = 0x3000 ;CMD_LOOP|Loop_Count (1 <= Loop_Count <= 256)
.equ	CMD_LOOPEND = 0x4000 ;CMD_LOOPEND
.equ	CMD_WAIT = 0x5000 ;CMD_WAIT|Wait_Time (1 <= Wait_Time <= 4095)

.dseg
RAM_CurrentFramePWM: .byte 16
RAM_CurrentFrame: .byte 16
RAM_OldFrame:  .byte 16
RAM_NewFrame:  .byte 16
RAM_RS232Buffer: .byte 32

;===============================================================================

.cseg
.org 0x000
	rjmp	start
.org URXC0addr
	rjmp	rs232

;===============================================================================

;ANIMATIONS

ani_list:
	.dw blinkenarea*2, congress22*2, ANI_END
	;.dw lauflicht*2, ANI_END
	
;--------------------

blinkenarea: ;displays "blinkenarea" and a smiley with crossfading (Arne Rossius 2005-10-20)
	.dw CMD_FADE|12, 0xFFFF, 0x0F0F, 0x0F0F, 0x00F0, 2 ;b
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0xFFFF, 0x000F, 0x000F, 0x000F, 2 ;L
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0x0000, 0xFFFF, 0x0000, 0x0000, 2 ;I
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0xFFFF, 0x0F00, 0x00F0, 0xFFFF, 2 ;N
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0xFFFF, 0x00F0, 0x0FF0, 0xF00F, 2 ;K/k
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0x0FF0, 0xF0FF, 0xFF0F, 0x0000, 2 ;e
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0xFFFF, 0x0F00, 0x00F0, 0xFFFF, 2 ;N
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0x0FFF, 0xF0F0, 0xF0F0, 0x0FFF, 2 ;A
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0xFFFF, 0xF0F0, 0xF0F0, 0x0F0F, 2 ;R
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0x0FF0, 0xF0FF, 0xFF0F, 0x0000, 2 ;e
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0x0FFF, 0xF0F0, 0xF0F0, 0x0FFF, 2 ;A
	.dw CMD_WAIT|100
	.dw CMD_FADE|12, 0x0000, 0x0000, 0x0000, 0x0000, 2
	.dw CMD_WAIT|16
	.dw CMD_FADE|12, 0xF0F0, 0x000F, 0x000F, 0xF0F0, 2 ; :)
	.dw CMD_WAIT|300
	.dw CMD_FADE|12, 0x0000, 0x0000, 0x0000, 0x0000, 2
	.dw CMD_WAIT|128
	.dw ANI_END

;--------------------

congress22: ;22C3 in flashing letters (Arne Rossius 2005-10-20)
	.dw CMD_SET|16, 0xF00F, 0xF0FF, 0x0F0F, 0x0000 ;2
	.dw CMD_FADE|16, 0x0000, 0x0000, 0x0000, 0x0000, 4
	.dw CMD_SET|16, 0xF00F, 0xF0FF, 0x0F0F, 0x0000 ;2
	.dw CMD_FADE|16, 0x0000, 0x0000, 0x0000, 0x0000, 4
	.dw CMD_SET|16, 0x0FF0, 0xF00F, 0xF00F, 0x0000 ;C
	.dw CMD_FADE|16, 0x0000, 0x0000, 0x0000, 0x0000, 4
	.dw CMD_SET|16, 0xF00F, 0xFF0F, 0xF0F0, 0x0000 ;3
	.dw CMD_FADE|16, 0x0000, 0x0000, 0x0000, 0x0000, 4
	.dw CMD_WAIT|256
	.dw ANI_END

;--------------------

fading_demo: ;some fading effects (Arne Rossius 2005-09-15/16)
	.dw CMD_FADE|8, 0xFFFF, 0x0000, 0x0000, 0x0000, 4
	.dw CMD_WAIT|64
	.dw CMD_FADE|8, 0x0000, 0xFFFF, 0x0000, 0x0000, 4
	.dw CMD_WAIT|64
	.dw CMD_FADE|8, 0x0000, 0x0000, 0xFFFF, 0x0000, 4
	.dw CMD_WAIT|64
	.dw CMD_FADE|8, 0x0000, 0x0000, 0x0000, 0xFFFF, 4
	.dw CMD_WAIT|64
	.dw CMD_FADE|8, 0x0000, 0x0000, 0xFFFF, 0xFFFF, 4
	.dw CMD_WAIT|64
	.dw CMD_FADE|8, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 4
	.dw CMD_WAIT|64
	.dw CMD_FADE|8, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 4
	.dw CMD_WAIT|128
	.dw CMD_FADE|16, 0x0000, 0x0000, 0x0000, 0x0000, 4
	.dw CMD_FADE|16, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 4
	.dw CMD_WAIT|128
	;the following lines are *very* complicated ;-)
	.dw CMD_FADE|5, 0xBFFF, 0xBFFF, 0xBFFF, 0xBFFF, 4
	.dw CMD_FADE|5, 0x7BFF, 0x7BFF, 0x7BFF, 0x7BFF, 4
	.dw CMD_FADE|5, 0x37BF, 0x37BF, 0x37BF, 0x37BF, 4
	.dw CMD_FADE|4, 0x048C, 0x048C, 0x048C, 0x048C, 4
	.dw CMD_FADE|5, 0x0048, 0x0048, 0x0048, 0x0048, 4
	.dw CMD_FADE|5, 0x0004, 0x0004, 0x0004, 0x0004, 4
	.dw CMD_FADE|5, 0x0000, 0x0000, 0x0000, 0x0000, 4
	.dw CMD_WAIT|128
	.dw ANI_END

;--------------------

hard_blink: ;"hard" blinking (Arne Rossius 2005-09-15/16)
	.dw CMD_LOOP|3
		.dw CMD_SET|64, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
		.dw CMD_SET|64, 0x0000, 0x0000, 0x0000, 0x0000
	.dw CMD_LOOPEND
	.dw ANI_END
	
;--------------------

lauflicht: ;Arne Rossius 2005-12-02
	.dw CMD_SET|64, 0xF000, 0x0000, 0x0000, 0x0000
	.dw CMD_SET|64, 0x0F00, 0x0000, 0x0000, 0x0000
	.dw CMD_SET|64, 0x00F0, 0x0000, 0x0000, 0x0000
	.dw CMD_SET|64, 0x000F, 0x0000, 0x0000, 0x0000
	.dw ANI_END

;===============================================================================

;INIT

start:
	;setup stackpointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;setup ports
	ldi	temp, 0xFF
	out	DDRB, temp
	ldi	temp, 0xC3
	out	PORTB, temp
	ldi	temp, 0x00
	out	DDRD, temp
	ldi	temp, 0x7F
	out	PORTD, temp
	
	;setup RS232
	ldi	temp, 1<<RXCIE | 1<<RXEN
	out	UCSRB, temp
	ldi	temp, FREQ/(16*BAUD-1)-1
	out	UBRRL, temp
	
	;initialize registers
	clr	hour
	clr	minute
	clr	second
	clr	YH ;never changed!
	clr	keycnt1
	clr	keycnt2
	
;===============================================================================

;ANIMATION (default mode)

ani:
	ldi	mode, MODE_ANI
	;initialize pointers to first animation
	ldi	aninum, LOW(ani_list*2) ;is near the beginning of the program, so address is <256
	rcall	load_ani

ani_loop:
	;check keys
	cpi	keycnt1, KEY_LONG
	brlo	PC+2
	rjmp	stream
	cpi	keycnt2, KEY_SHORT
	brlo	PC+2
	rjmp	clock
	
	;get and execute one animation command
	lpm	ani1L, Z+
	lpm	ani1H, Z+
	mov	temp, ani1H
	andi	ani1H, 0x0F
	andi	temp, 0xF0
	cpi	temp, HIGH(ANI_END)
		brne	cmd1
		rcall	load_ani
		rjmp	ani_loop
cmd1:
	cpi	temp, HIGH(CMD_SET)
		brne	cmd2
		ldi	YL, RAM_CurrentFrame
		rcall	load_frame
		rcall	display_frame
		rjmp	ani_loop
cmd2:
	cpi	temp, HIGH(CMD_FADE)
		brne	cmd3
		;backup current frame
		ldi	YL, RAM_CurrentFrame
		ldi	temp2, 16
	cmd2_backuploop:
		ld	temp, Y
		std	Y+(RAM_OldFrame-RAM_CurrentFrame), temp
		inc	YL
		dec	temp2
		brne	cmd2_backuploop
		;load the new frame (displayed after the fading)
		ldi	YL, RAM_NewFrame
		rcall	load_frame
		;load the step wait time
		lpm	ani2L, Z+
		lpm	ani2H, Z+
		;fade loop
		mov	anisteps, ani1L
		dec	anisteps
		clr	anistep
		inc	anistep
	cmd2_loop:
		;fade calc loop
		ldi	YL, RAM_CurrentFrame
	cmd2_calcloop:
		ldd	aniold, Y+(RAM_OldFrame-RAM_CurrentFrame)
		ldd	aninew, Y+(RAM_NewFrame-RAM_CurrentFrame)
		sub	aninew, aniold
		;negate if result is negative
		clt
		brpl	cmd2_positive
		neg	aninew
		set
	cmd2_positive:
		;multiplication: ani1L = anistep * aninew
		mov	temp2, anistep
		clr	ani1L
		;clr	ani1H
	cmd2_mulloop:
		add	ani1L, aninew
		;brcc	PC+2
		;inc	ani1H
		dec	temp2
		brne	cmd2_mulloop
		;division: aninew = ani1L / anisteps
		;ldi	temp2, 0
		clr	aninew
	cmd2_divloop:
		cp	ani1L, anisteps
		;cpc	ani1H, temp2
		brlo	cmd2_divend
		sub	ani1L, anisteps
		;sbci	ani1H, 0
		inc	aninew
		rjmp	cmd2_divloop
	cmd2_divend:
		lsl	ani1L ;round result
		cp	ani1L, anisteps
		brlo	PC+2
		inc	aninew
		;negate result if negated at the beginning
		brtc	PC+2
		neg	aninew
		add	aninew, aniold
		st	Y+, aninew
		;end fade calc loop
		sbrs	YL, 7 ;YL >= RAM_CurrentFrame+16 (0x80) ?
		rjmp	cmd2_calcloop
		;display calculated frame
		mov	ani1L, ani2L
		mov	ani1H, ani2H
		rcall	display_frame
		;end fade loop
		inc	anistep
		cp	anisteps, anistep
		brsh	cmd2_loop
		rjmp	ani_loop
cmd3:
	cpi	temp, HIGH(CMD_LOOP)
		brne	cmd4
		mov	aniloopL, ZL
		mov	aniloopH, ZH
		mov	aniloops, ani1L
		rjmp	ani_loop
cmd4:
	cpi	temp, HIGH(CMD_LOOPEND)
		brne	cmd5
		dec	aniloops
		breq	cmd4_end
		mov	ZL, aniloopL
		mov	ZH, aniloopH
	cmd4_end:
		rjmp	ani_loop
cmd5:
	cpi	temp, HIGH(CMD_WAIT)
		brne	cmd6
		rcall	display_frame
		;[rjmp	ani_loop]
cmd6:
	;unknown command
	rjmp	ani_loop

;--------------------
	
load_ani:
	;set the Z pointer to the start of the next animation
	mov	ZL, aninum
	ldi	ZH, 0
	lpm	ani1L, Z+
	lpm	ani1H, Z+
	subi	aninum, -2
	cpi	ani1L, ANI_END
	cpc	ani1H, ani1L ;(ANI_END is 0)
	breq	load_ani_restart
	movw	ZH:ZL, ani1H:ani1L
	ret
load_ani_restart:
	ldi	aninum, LOW(ani_list*2)
	rjmp	load_ani

;--------------------

load_frame:
	;load a frame into RAM[YL..YL+15]
	ldi	temp2, 8
load_frame_loop:
	lpm	temp, Z
	andi	temp, 0x0F
	st	Y+, temp
	lpm	temp, Z+
	swap	temp
	andi	temp, 0x0F
	st	Y+, temp
	dec	temp2
	brne	load_frame_loop
	ret

;===============================================================================

;STREAM

stream:
	ldi	mode, MODE_STREAM_WAIT4RELEASE
	
	;display stream logo until the first frame is received
	ldi	YL, RAM_CurrentFrame
	ldi	ZL, LOW(stream_logo*2)
	ldi	ZH, HIGH(stream_logo*2)
	rcall	load_frame
	
stream_loop:
	tst	keycnt1
	brne	PC+2
	ldi	mode, MODE_STREAM
	
	;poll keys:
	cli
	cpi	mode, MODE_STREAM_WAIT4RELEASE
	breq	stream_key2
	cpi	keycnt1, KEY_SHORT
	brlo	PC+2
	rjmp	ani
stream_key2:
	cpi	keycnt2, KEY_SHORT
	brlo	PC+2
	rjmp	clock
	sei
	
	;display frame
	ldi	ani1L, 1
	ldi	ani1H, 0
	rcall	display_frame
	
	rjmp	stream_loop
	
;--------------------

stream_logo: ;stream logo (right arrow)
	.dw	0x000F, 0xF0F0, 0xFF00, 0xFFF0
	
;--------------------

rs232:
	;RS232 Rx Complete interrupt
	push	temp
	push	temp2
	push	YL
	push	ZL
	push	ZH
	in	temp, SREG
	push	temp
	
	;reset timeout
	ldi	temp, RS232_TIMEOUT
	mov	rs232cnt, temp
	
	;write new character to buffer
	in	temp, UDR
	mov	YL, p_rs232
	st	Y+, temp
	inc	p_rs232
	
	;check for valid frame length (header: 12 Byte, data: 16 Byte)
	cpi	YL, RAM_RS232Buffer+12+16
	brlo	rs232_end
	breq	rs232_check
	;frame is too long:
rs232_error:
	;erroneous data received: receive no more chars until timeout
	ldi	p_rs232, RAM_RS232Buffer+30
	rjmp	rs232_end

rs232_check:
	;check for valid frame header (11 bytes, 12th byte is maxval)
	ldi	YL, RAM_RS232Buffer
	ldi	ZL, LOW(mcuf_header*2)
	ldi	ZH, HIGH(mcuf_header*2)
rs232_check_loop:
	ld	temp, Y+
	lpm	temp2, Z+
	cp	temp, temp2
	brne	rs232_error
	cpi	YL, RAM_RS232Buffer+11
	brlo	rs232_check_loop
	;header ok, get maxval data (must be 0x0F or 0xFF)
	ld	temp, Y+ ;LOW(maxval)
	cpi	temp, 0x0F
	breq	rs232_ok
	cpi	temp, 0xFF
	brne	rs232_error

rs232_ok:
	;copy frame data to currently displayed frame area
	;note that pixel order in MCUF and in RAM differs (conversion neccessary!)
	ldi	YL, RAM_RS232Buffer+12 ;12 byte offset due to header
	ldi	ZL, RAM_CurrentFrame+0x3 ;this register counts: 3,7,B,F,2,6,A,E,1,5,9,D,0,4,8,C
	ldi	ZH, 0
rs232_copy:
	ld	temp2, Y+
	cpi	temp, 0xFF ;if maxval = 0xFF: data = data >> 4
	brne	rs232_copy_noswap
	swap	temp2
	andi	temp2, 0x0F
rs232_copy_noswap:
	st	Z, temp2
	subi	ZL, -4
	cpi	ZL, RAM_CurrentFrame+16
	brlo	rs232_copy
	subi	ZL, 0x11
	cpi	ZL, RAM_CurrentFrame
	brsh	rs232_copy
	;reset RS232 RAM pointer to 0
	ldi	p_rs232, RAM_RS232Buffer
	
rs232_end:
	;return to main program
	pop	temp
	out	SREG, temp
	pop	ZH
	pop	ZL
	pop	YL
	pop	temp2
	pop	temp
	reti
	
;--------------------

mcuf_header:
	.db 0x23, 0x54, 0x26, 0x66 ;magic
	.db 0x00, 0x04, 0x00, 0x04 ;H=4, W=4
	.db 0x00, 0x01, 0x00, 0    ;1 Channel, maxval<0x0100, "0" is for padding
	
;===============================================================================
	
;CLOCK

clock:
	ldi	mode, MODE_CLOCK
	rcall	clock_update
	
clock_loop:
	;display clock
	ldi	ani1L, 16
	ldi	ani1H, 0
	rcall	display_frame
	;check left key
	cpi	keycnt1, KEY_SHORT
	brlo	PC+2
	rjmp	ani
	;check right key
	cpi	keycnt2, KEY_LONG
	brsh	clock_set
	rjmp	clock_loop
	
;--------------------
	
clock_inc:
	;increment clock
	inc	second
	cpi	second, 60
	brlo	clock_end
	;new minute
	clr	second
	inc	minute
	subi	minute, -6 ;add 6 (halfcarry set if NO overflow!)
	brhc	PC+2
	subi	minute, 6
	cpi	minute, 0x60
	brlo	clock_update
	;new hour
	clr	minute
	inc	hour
	subi	hour, -6 ;add 6 (halfcarry set if NO overflow!)
	brhc	PC+2
	subi	hour, 6
	cpi	hour, 0x24
	brlo	clock_update
	;end of day
	clr	hour
	
clock_update:
; 	;update BCD clock display if mode is MODE_CLOCK* (mode >= 4)
	cpi	mode, MODE_CLOCK
	brlo	clock_end
	ldi	YL, RAM_CurrentFrame
	mov	temp, hour
	swap	temp
	rcall	bcd2ram
	mov	temp, hour
	rcall	bcd2ram
	mov	temp, minute
	swap	temp
	rcall	bcd2ram
	mov	temp, minute
	rcall	bcd2ram
clock_end:
	ret
	
;-------------------

clock_set:
	ldi	mode, MODE_CLOCKSET_WAIT4RELEASE
	ldi	YL, RAM_CurrentFrame+3*4
	
clockset_blinkoff:
	;don't blink if left key (+1) pressed
	sbis	PIND, 2 ;left key
	rjmp	clockset_blinkoff_end
	;jump back to clock if all "digits" have been set
	cpi	YL, RAM_CurrentFrame
	brsh	PC+2
	rjmp	clock
	;blink off
	ldi	temp, 0
	ldi	temp2, 4
clockset_blinkoff_loop:
	st	Y+, temp
	dec	temp2
	brne	clockset_blinkoff_loop
	subi	YL, 4
	ldi	ani1L, 32
	ldi	ani1H, 0
	push	YL
	rcall	display_frame
	pop	YL
	rcall	clockset_keys
clockset_blinkoff_end:
	;blink on
	push	YL
	rcall	clock_update
	ldi	ani1L, 32
	ldi	ani1H, 0
	rcall	display_frame
	pop	YL
	rcall	clockset_keys
	rjmp	clockset_blinkoff
	
clockset_keys:
	;poll keys
	tst	keycnt2
	brne	clockkeys_start
	ldi	mode, MODE_CLOCKSET
clockkeys_start:
	cpi	mode, MODE_CLOCKSET
	brlo	clockkeys_end
	;long press
	cpi	keycnt1, KEY_LONG
	brlo	clockkeys_short
	rcall	clockset_inc
	rjmp	clockkeys_end
clockkeys_short:
	;short press
	cpi	keycnt1, KEY_SHORT
	brne	clockkeys_short2
	rcall	clockset_inc
	rjmp	clockkeys_end
clockkeys_short2:
	cpi	keycnt2, KEY_SHORT
	brne	clockkeys_end
	subi	YL, 4
clockkeys_end:
	ret
	
clockset_inc:
	;increment selected "digit"
	cpi	YL, RAM_CurrentFrame+0*4
	breq	clockset_inc_hr10
	cpi	YL, RAM_CurrentFrame+1*4
	breq	clockset_inc_hr1
	cpi	YL, RAM_CurrentFrame+2*4
	breq	clockset_inc_min10
	;minute 1s
	clr	second
	inc	minute
	mov	temp, minute
	andi	temp, 0x0F
	cpi	temp, 10
	brlo	PC+2
	andi	minute, 0xF0
	ret
clockset_inc_min10:
	;minute 10s
	subi	minute, -0x10
	cpi	minute, 0x60
	brlo	PC+2
	andi	minute, 0x0F
	ret
clockset_inc_hr1:
	;hour 1s
	inc	hour
	mov	temp, hour
	andi	temp, 0x0F
	cpi	temp, 10
	brlo	PC+2
	andi	hour, 0xF0
	ret
clockset_inc_hr10:
	;hour 10s
	subi	hour, -0x10
	cpi	hour, 0x24
	brlo	PC+2
	andi	hour, 0x0F
	ret

;--------------------

bcd2ram:
	;copy the BCD number bits at temp[3..0] to 4 consecutive RAM bytes (0=0, 1=15)
	andi	temp, 0x0F
	cpi	mode, MODE_CLOCKSET_WAIT4RELEASE
	brlo	bcd2ram_noset
	cpi	temp, 0
	brne	bcd2ram_noset
	ldi	temp, 15 ;display 0000 as 1111 while setting the clock
bcd2ram_noset:
	ldi	temp2, 0x01
bcd2ram_loop:
	push	temp
	and	temp, temp2
	breq	PC+2
	ldi	temp, 15
	st	Y+, temp
	pop	temp
	lsl	temp2
	brhc	bcd2ram_loop
	ret

;===============================================================================

;DISPLAY

display_frame:
	;display the current frame ani1H:ani1L times
	ldi	YL, RAM_CurrentFramePWM
	ldi	temp2, 16
	;convert linear X=[0..F] to approx. 2^(X/15*8)-1
	push	ZL
	push	ZH
convert_loop:
	ldd	temp, Y+(RAM_CurrentFrame-RAM_CurrentFramePWM)
	ldi	ZL, LOW(lin2exp_val*2)
	ldi	ZH, HIGH(lin2exp_val*2)
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm	temp, Z
	st	Y+, temp
	dec	temp2
	brne	convert_loop
	pop	ZH
	pop	ZL
display_loop:
	;display frame
	ldi	temp, 255
	mov	pwmcnt, temp
pwm_loop:
	;PWM loop
	ldi	temp, 0x20
	mov	mpxcnt, temp
	ldi	YL, RAM_CurrentFramePWM
mpx_loop:
	;Multiplex loop (multiplex all 4 lines with PWM)
	mov	temp, mpxcnt
	ld	temp2, Y+	;LED 1
	cp	temp2, pwmcnt
	brsh	PC+2
	ori	temp, 0x80
	ld	temp2, Y+	;LED 2
	cp	temp2, pwmcnt
	brsh	PC+2
	ori	temp, 0x40
	ld	temp2, Y+	;LED 3
	cp	temp2, pwmcnt
	brsh	PC+2
	ori	temp, 0x02
	ld	temp2, Y+	;LED 4
	cp	temp2, pwmcnt
	brsh	PC+2
	ori	temp, 0x01
	out	PORTB, temp
	;short delay
	ldi	temp2, 8
pwm_wait_loop:
	dec	temp2
	brne	pwm_wait_loop
	
	andi	temp, 0xC3 ;clear Enable lines
	out	PORTB, temp
	
	;end multiplex loop
	;subi	YL, -4
	lsr	mpxcnt
	sbrs	mpxcnt, 1 ;mpxcnt < 0x04 ?
	rjmp	mpx_loop
	
	;end PWM loop
	;subi	YL, 4
	subi	YL, 16
	dec	pwmcnt
	brne	pwm_loop
	
	;check clock input (1 pulse per second)
	in	temp, PIND
	andi	temp, 0x10
	cp	temp, clkin
	breq	clkin_end
	mov	clkin, temp
	.if ClockPulses != 0
		sbrs	temp, 4 ;increment clock at falling edge
	.else
		sbrc	temp, 4 ;increment clock at rising edge
	.endif
	rcall	clock_inc
clkin_end:

	;check keys
	inc	keycnt1 ;left key
	sbic	PIND, 2
	clr	keycnt1
	inc	keycnt2 ;right key
	sbic	PIND, 3
	clr	keycnt2
	;long press
	cpi	keycnt1, KEY_LONG
	brlo	PC+2
	ldi	keycnt1, KEY_LONG
	cpi	keycnt2, KEY_LONG
	brlo	PC+2
	ldi	keycnt2, KEY_LONG
	;short press
	cpi	keycnt1, KEY_SHORT
	breq	display_end
	cpi	keycnt2, KEY_SHORT
	breq	display_end
	
	;RS232 timeout counter
	dec	rs232cnt
	brne	PC+2
	ldi	p_rs232, RAM_RS232Buffer
	
	;end display frame loop
	sbiw	ani1H:ani1L, 1
	breq	display_end
	rjmp	display_loop
display_end:
	ret

;--------------------
	
lin2exp_val: ;PWM values for the 16 different brightness values
	.db 0, 1, 2, 3, 4, 6, 8, 12, 18, 27, 39, 57, 83, 121, 176, 255
